@@ -45,6 +45,7 @@ OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMA
 ##
 ##      Windowgram Convert
 ##      WindowgramGroup Convert
+##      Flex Cores
 ##      Flex Modifiers
 ##      Readme Demonstrations
 ##
@@ -120,15 +121,21 @@ def UnitTests():
 class SenseTestCase(unittest.TestCase):
 
     def runTest(self):
-        # Sense test methods
-        methods = inspect.getmembers(self.__class__(), predicate=inspect.ismethod)
-        methods = [ method for method in methods if method[0].startswith("test_") ]
-        # Pair with line numbers
-        methods = [ (method[1], inspect.getsourcelines(method[1])[1]) for method in methods ]
-        # Sort by line numbers
-        methods = sorted(methods, key=lambda tup: tup[1])
-        # Execute tests in the order they appear
-        for method in methods: method[0]()
+        try:
+            # Sense test methods
+            methods = inspect.getmembers(self.__class__(), predicate=inspect.ismethod)
+            methods = [ method for method in methods if method[0].startswith("test_") ]
+            # Pair with line numbers
+            methods = [ (method[1], inspect.getsourcelines(method[1])[1]) for method in methods ]
+            # Sort by line numbers
+            methods = sorted(methods, key=lambda tup: tup[1])
+            # Execute tests in the order they appear
+            for method in methods: method[0]()
+        except AssertionError as e:
+            raise e # Forward
+        except Exception as e:
+            error = "An error occurred during testing: " + repr(e) # Show the error in case of failure during test
+            raise AssertionError( e ) # Forward the exception to properly halt because of failure
 
     ##----------------------------------------------------------------------------------------------------------
     ##
@@ -266,6 +273,17 @@ class Test_Windowgram_Convert(SenseTestCase):
         data_x = Windowgram_Convert.Purify( data_i )
         self.assertTrue( data_x == data_o )
 
+    def test_Windowgram_Convert_TransposeCharacters(self):
+        data_i = [  [ "1", "1", "3", "5" ],
+                    [ "1", "1", "4", "5" ],
+                    [ "2", "2", "4", "5" ] ]
+        data_o = [  [ "1", "1", "2" ],
+                    [ "1", "1", "2" ],
+                    [ "3", "4", "4" ],
+                    [ "5", "5", "5" ] ]
+        data_x = Windowgram_Convert.Transpose_Chars( data_i )
+        self.assertTrue( data_x == data_o )
+
 
 
 ##----------------------------------------------------------------------------------------------------------------------
@@ -372,17 +390,324 @@ class Test_WindowgramGroup_Convert(SenseTestCase):
 
 ##----------------------------------------------------------------------------------------------------------------------
 ##
-## Unit Testing :: Flex Modifiers
+## Unit Testing :: Flex Cores
 ##
 ##----------------------------------------------------------------------------------------------------------------------
 
-class Test_FlexModifiers(SenseTestCase):
+class Test_FlexCores(SenseTestCase):
 
-    ##----------------------------------------------------------------------------------------------------------
-    ##
-    ## Flex Modifier: Scale
-    ##
-    ##----------------------------------------------------------------------------------------------------------
+    ## Scale Core
+
+    def test_ScaleCore(self):
+        group_i = """
+            111223
+            111223
+            111223
+            xxxyyz
+            xxxyyz
+            XXXYYZ
+        """
+        group_o = Windowgram_Convert.Lines_To_String( [ "111111222233", "111111222233", "xxxxxxyyyyzz" ] )
+        group_x = scalecore( group_i, 12, 3 )
+        self.assertTrue( group_o == group_x )
+
+    ## Group Core
+
+    def test_GroupCore_Sufficient(self):
+        wg = Windowgram( """
+            qqwwee
+            qqwwee
+            rrttyy
+            rrttyy
+            uuiioo
+            uuiioo
+        """ )
+        result, suggestions = groupcore( wg, "qwrt" )
+        self.assertTrue( result == GroupStatus.Success )
+        self.assertTrue( suggestions == "" )
+
+    def test_GroupCore_Insufficient(self):
+        wg = Windowgram( """
+            1122xx
+            1122yy
+            33zzzz
+            33zzzz
+            rrzzzz
+            rrsstt
+        """ )
+        result, suggestions = groupcore( wg, "123" )
+        self.assertTrue( result == GroupStatus.Insufficient_Panes )
+        self.assertTrue( suggestions == PaneList_AssimilatedSorted( "rstxyz", "" ) )
+
+    ## Edge Core
+
+    def test_EdgeCore_Merger(self):
+        #
+        # The merger algorithm should be obvious.  Required to merge the results of SwipeSide algorithm.
+        #
+        self.assertTrue( edgecore_merger( [[3, 3, 5], [9, 0, 2], [9, 3, 5], [9, 6, 8]] ) == [[3, 3, 5], [9, 0, 8]] )
+        self.assertTrue( edgecore_merger( [[9, 6, 8], [9, 3, 5], [3, 3, 5], [9, 0, 2]] ) == [[3, 3, 5], [9, 0, 8]] )
+
+    def test_EdgeCore_SideSwipe(self):
+        #
+        # SideSwipe algorithm (not to be confused with SwipeSide)
+        #
+        # This algorithm is axis-agnostic; illustrated tests are vertical, i.e., horizontal swaps parameters.
+        #
+        # To produce these tests, inject the respective windowgram into edgecore() and print the results then exit.
+        #
+        # Successful    Successful    Fail    Fail
+        # @A @B @C @D   @E @F @G @H   @I @J   @K @L
+        # .2 1. .2 1.   .1 2. .1 2.   .2 1.   2. .1
+        # .2 1. 12 12   .1 2. 21 21   .2 1.   2. ..
+        # 12 1. 12 12   21 2. 21 21   .2 ..   2. 2.
+        # 12 12 12 12   21 21 21 21   .2 ..   .1 2.
+        # 12 12 12 .2   21 21 21 .1   .2 .2   .1 2.
+        # 12 12 .2 .2   21 21 .1 .1   1. .2   .1 2.
+        # 1. 1. .2 .2   2. 2. .1 .1   1. .2   .1 2.
+        #                                    p1x   p2x   p1y   p2y
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 3, 7, 1, 6 ) == [ 1, 2, 5 ] )  # @A
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 1, 7, 4, 6 ) == [ 1, 3, 5 ] )  # @B
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 2, 5, 1, 7 ) == [ 1, 1, 4 ] )  # @C
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 1, 4, 2, 7 ) == [ 1, 1, 3 ] )  # @D
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 1, 6, 3, 7 ) == [ 1, 2, 5 ] )  # @E
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 4, 6, 1, 7 ) == [ 1, 3, 5 ] )  # @F
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 1, 7, 2, 5 ) == [ 1, 1, 4 ] )  # @G
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 2, 7, 1, 4 ) == [ 1, 1, 3 ] )  # @H
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 6, 7, 1, 5 ) == None )         # @I
+        self.assertTrue( edgecore_sideswipe( 1, 1, 2, 2, 1, 2, 5, 7 ) == None )         # @J
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 4, 7, 1, 3 ) == None )         # @K
+        self.assertTrue( edgecore_sideswipe( 2, 2, 1, 1, 1, 1, 3, 7 ) == None )         # @L
+
+    def test_EdgeCore_SwipeSide(self):
+        #
+        # SwipeSide algorithm (not to be confused with SideSwipe)
+        #
+        # This algorithm is axis-agnostic; when using X major, the windowgram and pane must be transposed.
+        #
+        wg = Windowgram( """
+            xx000yyyy # xxx111333 (original, transposed)
+            xx000yyyy # xxx111333
+            xx000yyyy # 000zzz444
+            11zzz2222 # 000zzz444
+            11zzz2222 # 000zzz444
+            11zzz2222 # yyy444555
+            334445555 # yyy444555
+            334445555 # yyy444555
+            334445555 # yyy444555
+            """ )
+        windowgram_chars_yx = wg.Export_Chars()
+        windowgram_chars_xy = Windowgram_Convert.Transpose_Chars( windowgram_chars_yx )
+        windowgram_parsed = wg.Export_Parsed()
+        group = "012345"
+        # For cleaner processing
+        def edgescan_q(axis, direction, windowgram_parsed, group, windowgram_chars):
+            runs = []
+            for pane in list(group):
+                parsedpane = windowgram_parsed[pane]
+                if not axis: parsedpane = Windowgram_Convert.Transpose_Pane( parsedpane )
+                this_runs = edgecore_swipeside( direction, parsedpane, group, windowgram_chars )
+                for run in this_runs: runs.append( run )
+            return edgecore_merger( runs )
+        def edgescan_v(direction):
+            return edgescan_q( True, direction, windowgram_parsed, group, windowgram_chars_yx )
+        def edgescan_h(direction):
+            return edgescan_q( False, direction, windowgram_parsed, group, windowgram_chars_xy )
+        # TBLR
+        self.assertTrue( edgescan_v( True )  == [[0, 2, 4], [3, 0, 1], [3, 5, 8], [6, 2, 4]] )
+        self.assertTrue( edgescan_v( False ) == [[3, 2, 4], [9, 0, 8]] )
+        self.assertTrue( edgescan_h( True )  == [[0, 3, 8], [2, 0, 2], [5, 3, 5]] )
+        self.assertTrue( edgescan_h( False ) == [[2, 3, 5], [5, 0, 2], [9, 3, 8]] )
+
+    def test_EdgeCore_Invalid(self):
+        wg = Windowgram( """
+            W11YY
+            XX22Z
+        """ )
+        # Irrational (W does not border Z)
+        status, axis, minimal, optimal = edgecore( wg, "WZ" )
+        self.assertTrue( status is EdgeStatus.Irrational )
+        self.assertTrue( axis is None )
+        self.assertTrue( minimal is None )
+        self.assertTrue( optimal is None )
+        # Ambiguous (vertical edge W1 and X, horizontal edge W and 1)
+        status, axis, minimal, optimal = edgecore( wg, "WX1" )
+        self.assertTrue( status is EdgeStatus.Ambiguous )
+        self.assertTrue( axis is None )
+        self.assertTrue( minimal is None )
+        self.assertTrue( optimal is None )
+        # Noncontiguous (a gap exists in the minimal edge between 1 and 2)
+        status, axis, minimal, optimal = edgecore( wg, "WXYZ" )
+        self.assertTrue( status is EdgeStatus.Noncontiguous )
+        self.assertTrue( axis is None )
+        self.assertTrue( minimal is None )
+        self.assertTrue( optimal is None )
+
+    def test_EdgeCore_Group_Vertical(self):
+        wg = Windowgram( """
+            WX
+            1X
+            12
+            Y2
+            YZ
+        """ )
+        status, axis, minimal, optimal = edgecore( wg, "12" ) # Implicit vertical
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis == "v" )
+        self.assertTrue( minimal == [ [1, 2, 3] ] )
+        self.assertTrue( optimal == [ [1, 0, 5] ] )
+
+    def test_EdgeCore_Group_Horizontal(self):
+        wg = Windowgram( """
+            1AA22
+            1AA22
+            33BB4
+            33BB4
+        """ )
+        status, axis, minimal, optimal = edgecore( wg, "AB" ) # Implicit horizontal
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis == "h" )
+        self.assertTrue( minimal == [ [2, 2, 3] ] )
+        self.assertTrue( optimal == [ [2, 0, 5] ] )
+
+    def test_EdgeCore_GroupDirection_Vertical(self):
+        wg = Windowgram( """
+            OWbo
+            O1Xo
+            O12o
+            OY2o
+            OqZo
+        """ )
+        status, axis, minimal, optimal = edgecore( wg, "2", "left" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis == "v" )
+        self.assertTrue( minimal == [ [2, 2, 4] ] )
+        self.assertTrue( optimal == [ [2, 1, 4] ] )
+
+    def test_EdgeCore_GroupDirection_Horizontal(self):
+        wg = Windowgram( """
+            1AA22
+            33BB4
+        """ )
+        status, axis, minimal, optimal = edgecore( wg, "A", "bottom" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis == "h" )
+        self.assertTrue( minimal == [ [1, 1, 3] ] )
+        self.assertTrue( optimal == [ [1, 0, 5] ] )
+
+    def test_EdgeCore_WindowgramEdge(self): # All edges here will be equal, there are no neighbors to extend optimal
+        wg = Windowgram( """
+            011222
+            344555
+            344555
+            677888
+            677888
+            677888
+        """ )
+        status, axis, minimal, optimal = edgecore( wg, "5", "right" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis is "v" )
+        self.assertTrue( minimal == [ [6, 1, 3] ] )
+        self.assertTrue( optimal == [ [6, 1, 3] ] )
+        status, axis, minimal, optimal = edgecore( wg, "0", "left" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis is "v" )
+        self.assertTrue( minimal == [ [0, 0, 1] ] )
+        self.assertTrue( optimal == [ [0, 0, 1] ] )
+        status, axis, minimal, optimal = edgecore( wg, "8", "bottom" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis is "h" )
+        self.assertTrue( minimal == [ [6, 3, 6] ] )
+        self.assertTrue( optimal == [ [6, 3, 6] ] )
+        status, axis, minimal, optimal = edgecore( wg, "1", "top" )
+        self.assertTrue( status is EdgeStatus.Valid )
+        self.assertTrue( axis is "h" )
+        self.assertTrue( minimal == [ [0, 1, 3] ] )
+        self.assertTrue( optimal == [ [0, 1, 3] ] )
+
+
+
+##----------------------------------------------------------------------------------------------------------------------
+##
+## Unit Testing :: Flex Modifier :: Scale
+##
+##----------------------------------------------------------------------------------------------------------------------
+
+class Test_FlexModifier_Scale(SenseTestCase):
+
+    ## Enforces the use of scale core v1, and will fail if v2 is reactivated.  See notes in scale core.
+
+    def test_Scale_ScaleCore_CaseOne(self): # Created in flex using "new unittest Scale_ScaleCore_CaseOne"
+        self.assertFlexSequence( [
+            "break 1 2x2 ; scale 3x3 ; scale 2x2",
+        ], """
+            01
+            23
+        """ )
+
+    ## Tests the scale core multiple resizing that's necessary in some situations.  See notes in scale core.
+
+    def test_Scale_ScaleCore_CaseTwo(self): # Created in flex using "new unittest Scale_ScaleCore_CaseTwo"
+        self.assertFlexSequence( [
+            "scale 42x42 ; break 1 6x6 ; break 1 3x3",
+        ], """
+            000000000111AAABBB222222222333333333444444444555555555
+            000000000111AAABBB222222222333333333444444444555555555
+            000000000111AAABBB222222222333333333444444444555555555
+            000000000CCCDDDEEE222222222333333333444444444555555555
+            000000000CCCDDDEEE222222222333333333444444444555555555
+            000000000CCCDDDEEE222222222333333333444444444555555555
+            000000000FFFGGGHHH222222222333333333444444444555555555
+            000000000FFFGGGHHH222222222333333333444444444555555555
+            000000000FFFGGGHHH222222222333333333444444444555555555
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            666666666777777777888888888999999999aaaaaaaaabbbbbbbbb
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            cccccccccdddddddddeeeeeeeeefffffffffggggggggghhhhhhhhh
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            iiiiiiiiijjjjjjjjjkkkkkkkkklllllllllmmmmmmmmmnnnnnnnnn
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            ooooooooopppppppppqqqqqqqqqrrrrrrrrrsssssssssttttttttt
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+            uuuuuuuuuvvvvvvvvvwwwwwwwwwxxxxxxxxxyyyyyyyyyzzzzzzzzz
+        """ )
+
+    ## General use
 
     def test_Scale_One_DupCharacters(self): # Created in flex using "new unittest Scale_One_DupCharacters"
         self.assertFlexSequence( [
@@ -547,15 +872,17 @@ class Test_FlexModifiers(SenseTestCase):
                                         11111111111111111111
         """ )
 
-    ##----------------------------------------------------------------------------------------------------------
-    ##
-    ## Keep this note for adding new unit tests for flex
-    ##
-    ##   flex> new unittest ScaleCommand    # When created, the output switches to a convenient code dump
-    ##   flex> scale 25x10                  # Run commands and the unittest code will be built as you go
-    ##   flex> scale 20x20                  # When finished just paste the generated code into this class
-    ##
-    ##----------------------------------------------------------------------------------------------------------
+
+
+##----------------------------------------------------------------------------------------------------------
+##
+## Keep this note for adding new unit tests for flex
+##
+##   flex> new unittest ScaleCommand    # When created, the output switches to a convenient code dump
+##   flex> scale 25x10                  # Run commands and the unittest code will be built as you go
+##   flex> scale 20x20                  # When finished just paste the generated code into this class
+##
+##----------------------------------------------------------------------------------------------------------
 
 
 
